import os


def exists(file_path: str):
    """
    Checks if the NMSLIB file exists and was saved with the data included.
    :param file_path: File path pointing to the NMSLIB index.
    :return: True if both nmslib index and their data exists, False otherwise.
    """
    index_exists: bool = os.path.isfile(file_path)
    index_data_exists: bool = os.path.isfile(f'{file_path}.dat')
    return index_exists and index_data_exists



import numpy as np
import matplotlib.pyplot as plt

from mpl_toolkits import mplot3d


def show(data_set, number_points: int):
    """
    Visualize data set plotting the points with a certain amount of points as maximum.
    :param data_set: Set of points to plot.
    :param number_points: Amount of points maximum to show.
    :return: Data set plotted.
    """
    print(f'info: Showing {number_points} as maximum.')
    sub_set_points = np.random.choice(range(data_set.shape[0]), size=min(data_set.shape[0], number_points))
    x = data_set[sub_set_points, 0]
    y = data_set[sub_set_points, 1]
    z = data_set[sub_set_points, 2]

    fig = plt.figure(figsize=(8, 8))
    ax = mplot3d.Axes3D(fig)
    ax.set_title('NMSLIB index 3D representation', fontsize=20)
    ax.scatter(xs=x, ys=y, zs=z)
    plt.show()



# Parameters
VIEW_DIMENSIONS: int = 3
NUMBER_POINTS: int = 1000

# Argument helper
HELP_NMSLIB_INDEX_FILE_PATH = 'String pointing to the NMSLIB file path. ' \
                              'There has to be the additional .dat file for the index, the one that is created when ' \
                              'uses the `save_data=True` option in the saveIndex() function. For example: ' \
                              '`index.nmslib` and `index.nmslib.dat` need to exist but only `index.nmslib` has to be ' \
                              'provided as this parameter.'
HELP_NUMBER_POINTS = 'Number of points as a maximum to plot from the NMSLIB index data, where has to be 1 at least. ' \
                     f'Default value: {NUMBER_POINTS}.'

# Error messages
ERROR_FILE_NOT_EXISTS = 'File path provided do not point to an NMSLIB index. ' \
                        'Remember for visualizing an NMSLIB index, data from the index had to be saved when saving ' \
                        'the index using the `save_data=True` option in the saveIndex() function.'
ERROR_NUMBER_POINTS = 'Number of points has to be at least 1.'



from nmslib_viz import file_manager, nmslib, pca, plot
from nmslib_viz.config import ERROR_FILE_NOT_EXISTS, ERROR_NUMBER_POINTS, NUMBER_POINTS


def view(nmslib_index_file_path: str, number_points: int = NUMBER_POINTS):
    """
    Main function for visualizing an NMSLIB index given its file path.
    :param nmslib_index_file_path: File path pointing to a NMSLIB index.
    :param number_points: Number of points as a maximum to plot from the NMSLIB index data.
    :return: Index being visualized.
    """
    try:
        # Check if the file exists
        file_exists: bool = file_manager.exists(nmslib_index_file_path)
        if not file_exists:
            print(f'error: {ERROR_FILE_NOT_EXISTS}')
            return
        # Check number of points
        if number_points <= 0:
            print(f'error: {ERROR_NUMBER_POINTS}')
            return
        # Load index
        print('info: Loading index...')
        index_instance = nmslib.load_index(nmslib_index_file_path)
        if file_exists is None:
            return
        # Build data set
        print('info: Building data set...')
        data_set = nmslib.build_data_set(index_instance)
        # PCA fit-transform
        print('info: PCA fit-transforming...')
        data_set = pca.reduce(data_set)
        if data_set is None:
            return
        # Plot
        print('info: Plotting data...')
        plot.show(data_set, number_points)
    except Exception as e:
        print(f'error: Unexpected error occurred: [{e}]')
    finally:
        print('info: Done!')



import numpy as np

from sklearn.decomposition import PCA

from nmslib_viz.config import VIEW_DIMENSIONS


def reduce(data_set):
    """
    Reduce data set dimensionality to the use set up on the configuration.
    :param data_set: Numpy array representing the data set.
    :return: Data set dimensionality reduce to the desired dimensions.
    """
    try:
        print(f'info: reducing features from {data_set.shape[1]} to {VIEW_DIMENSIONS} dimensions...')
        pca_instance = PCA(n_components=VIEW_DIMENSIONS, svd_solver='full')
        # If number of samples is lower than dimensionality
        samples_amount: int = data_set.shape[0]
        if samples_amount < VIEW_DIMENSIONS:
            new_data_set = np.empty((VIEW_DIMENSIONS, data_set.shape[1]), dtype=data_set.dtype)
            new_data_set[:samples_amount] = data_set
            new_data_set[samples_amount:] = data_set[np.random.choice(range(samples_amount),
                                                                      size=VIEW_DIMENSIONS - samples_amount)]
            pca_instance.fit(new_data_set)
        else:
            pca_instance.fit(data_set)
        # Reducing dimensions
        features = pca_instance.transform(data_set)
        return features
    except Exception as e:
        print(f'error: Data set could not be dimensionality reduced to {VIEW_DIMENSIONS}: [{e}]')
        return None



import setuptools


with open('README.md', 'r') as readme_file:
    long_description = readme_file.read()

setuptools.setup(
    name='nmslib-viz',
    version='1.0.0',
    author='AlbertSuarez',
    author_email='alsumo95@gmail.com',
    description='NMSLIB visualization tool',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/AlbertSuarez/nmslib-viz',
    license='Apache-2.0',
    packages=['nmslib_viz'],
    install_requires=[
        'nmslib==2.0.5',
        'numpy==1.18.1',
        'matplotlib==3.2.0',
        'scikit-learn==0.22.2.post1'
    ],
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    keywords='nmslib visualization graph',
    scripts=['bin/nmslib-viz']
 )



import nmslib
import numpy as np

from nmslib_viz.config import ERROR_FILE_NOT_EXISTS


def load_index(file_path: str):
    """
    Load an NMSLIB index with the data included given their file path.
    :param file_path: File path pointing to the NMSLIB index.
    :return: NMSLIB index instance.
    """
    try:
        index_instance = nmslib.init()
        index_instance.loadIndex(file_path, load_data=True)
        return index_instance
    except Exception as e:
        print(f'error: {ERROR_FILE_NOT_EXISTS} - {e}')
        return None


def build_data_set(index_instance):
    """
    Build a Numpy array with all the points inside the given index instance.
    :param index_instance: NMSLIB index instance.
    :return: Numpy array with all the points representing the data set.
    """
    data_set: list = list()
    nodes_amount: int = nmslib.getDataPointQty(index_instance)
    print(f'info: {nodes_amount} nodes found.')
    for idx in range(nodes_amount):
        try:
            data_set.append(nmslib.getDataPoint(index_instance, idx))
        except Exception as e:
            print(f'warn: Node number {idx} could not be obtained: [{e}]')
    return np.array(data_set)
